"""
Started versioning very late, as of 20210425

analysisVersion: saved for each spike
				indicates the analysis we used
				update this file as I change the analysis
"""

# 20210425
# analysisVersion = '20210426a' # backend engine
# interfaceVersion = '20210426a' # pyqt interface

# Laura returned from Italy, getting her most recent version
# analysisVersion = '20210803a' # backend engine
analysisVersion = "20230324a"  # backend engine

interfaceVersion = "20230324a"  # pyqt interface

# manually keep in sync with setup.py
__version__ = "0.1.11"